function [Hc, gc0, gc1, lbU, ubU, Ac, lbA0, ubA0, bA1, AA, BB, CC] = condensing(Q, R, P, g, A, B, lb, ub, beq)
%  This function is provided as part of the "TEMPO Summer School on 
%  Numerical Optimal Control and Embedded Optimization". It implements 
%  condensing in a rather simple way, such that it should be used merely
%  in an offline fashion to implement linear MPC in combination with qpOASES.
%
%  Course url: http://syscop.de/teaching/numerical-optimal-control/
%
%  NOTE: This is definitely NOT the most efficient algorithm to implement
%  condensing. The code should therefore be used ONLY for illustrative
%  purposes and/or to perform offline computations.
%
%    Author: Dimitris Kouzoupis <dimitris.kouzoupis@imtek.uni-freiburg.de>
%    Date: 2015
% 

nx = size(Q,1);
nu = size(R,1);
N = (length(g)-nx)/(nx+nu);

if length(beq) ~= N*nx
    error('Not supported: you should not add the initial value condition to the equality constraints!');
end
if lb(1:nx) == ub(1:nx)
    error('Not supported: you should not pass the initial state value to the condensing routine!');
end

% split bounds on state and input constraints

lb  = [lb; nan*ones(nu,1)];
ub  = [ub; nan*ones(nu,1)];

lb  = reshape(lb,nx+nu,N+1);
ub  = reshape(ub,nx+nu,N+1);

lbX = lb(1:end-nu,:);
lbX = lbX(:);
ubX = ub(1:end-nu,:);
ubX = ubX(:);
lbU = lb(end-nu+1,:);
ubU = ub(end-nu+1,:);
lbU = lbU(1:end-1)';
ubU = ubU(1:end-1)';

% extract system dynamics and weights

c = zeros(nx,1,N);
q = zeros(nx,1,N+1);
r = zeros(nu,1,N);

for ii = 1:N
    
    c(:,:,ii) = -beq((ii-1)*nx+1:ii*nx);
    
    q(:,:,ii) = g((ii-1)*(nx+nu)+1:(ii-1)*(nx+nu)+nx);
    r(:,:,ii) = g((ii-1)*(nx+nu)+nx+1:ii*(nx+nu));

end

% express states as: X = AA*x0 + BB*U + CC

% compute AA

Aii = eye(nx);
AA  = zeros((N+1)*nx,nx);
for ii = 1:N
    AA((ii-1)*nx+1:ii*nx,:) = Aii;
    Aii = A*Aii;
end

AA(N*nx+1:(N+1)*nx,:) = Aii;

% compute BB

BB = zeros((N+1)*nx,N*nu);
for ii = 1:N % block columns
    BB(ii*nx+1:(ii+1)*nx,(ii-1)*nu+1:ii*nu) = B;
    for jj = 1:N-ii % block rows
        BB(ii*nx+1+jj*nx:(ii+1)*nx+jj*nx,(ii-1)*nu+1:ii*nu) = ...
            A*BB(ii*nx+1+(jj-1)*nx:(ii+1)*nx+(jj-1)*nx,(ii-1)*nu+1:ii*nu);
    end
end

% compute CC

CC = zeros(N*nx,1);
for ii = 1:N
    CC(ii*nx+1:(ii+1)*nx) = A*CC((ii-1)*nx+1:ii*nx) + c(:,:,ii);
end

% concatinate Q and R matrices

QQ = [];
RR = [];
for ii = 1:N
    QQ = blkdiag(QQ,Q);
    RR = blkdiag(RR,R);
end
QQ = blkdiag(QQ,P);

% build condensed hessian and gradient

Hc = BB'*QQ*BB + RR;
grad0 = (BB'*QQ*AA);
gc0 = BB'*QQ*CC + 2*r(:) + 2*BB'*q(:);
gc1 = grad0;

% build polyhedral constraints
        
Ac = []; lbA0 = []; ubA0 = []; bA1 = [];
for ii = 1:length(lbX)
    if lbX(ii) > -1e10 || ubX(ii) < 1e10
        lbA0 = [lbA0; lbX(ii) - CC(ii)];
        ubA0 = [ubA0; ubX(ii) - CC(ii)];
        Ac = [Ac; BB(ii,:)];
        bA1 = [bA1; -AA(ii,:)];
    end
end

end

